#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "/usr/local/pkgs/matlab-r2006a/extern/include/mex.h"

#define BIG_ENDIAN_CPU		0
#define LITTLE_ENDIAN_CPU	1
#define BIG_ENDIAN_FILE		BIG_ENDIAN_CPU
#define LITTLE_ENDIAN_FILE	LITTLE_ENDIAN_CPU


void decomp_mayo(char *f_name, unsigned long long int start_idx, unsigned long long int end_idx, int *decomp_arr)
{
	char			*c, keysample_flag, *comp_data, *cdp, *data_end, *cvp, *b1, *b2, t;
	short			temp_ks_flag;
	unsigned int		cpu_endianness, file_endianness;
	unsigned long long int	j, time_file_len, *time_data, temp_ui8, start_data_file_offset, end_data_file_offset;
	unsigned long long int	n_read, curr_idx, comp_data_len;
	int			i, k, n_time_entries, tot_time_entries, f_name_len, current_val;
	FILE			*fp;

	/* Get endianness of local computer */
	cpu_endianness = 1;
	c = (char *) &cpu_endianness;
	if (*c)
		cpu_endianness = BIG_ENDIAN_CPU;
	else
		cpu_endianness = LITTLE_ENDIAN_CPU;
		
	/* Open associated time file */
	f_name_len = strlen(f_name);
	f_name[f_name_len - 2] = 't';			/* temporarily change "mef" to mtf" */
	fp = fopen(f_name, "r");
	if (fp == NULL) { 
		printf("Could not open the time file \"%s\" --> exiting\n",  f_name);
		return;
	}
	f_name[f_name_len - 2] = 'e';			/* change "mtf" back to mef" */
	
	/* Get file endianness */
	fread((void *) &file_endianness, sizeof(char), (size_t) 1, fp);
	if (file_endianness)
		file_endianness = LITTLE_ENDIAN_FILE;
	else
		file_endianness = BIG_ENDIAN_FILE;
	
	/* process seperately for endianness match & mismatch */
	if (cpu_endianness == file_endianness) {
		
		/* read in time file */
		fseek(fp, 16, SEEK_SET);		/* number of time entries offset */
		fread(&temp_ui8, sizeof(unsigned long long int), 1, fp);
		n_time_entries = (int) temp_ui8;	/* if exceeds this, to big for Matlab anyway */
		tot_time_entries = n_time_entries * 3;	/* 3 fields per entry */
		fseek(fp, 0, SEEK_END);			/* get file length */
		time_file_len = ftell(fp) - 1024;
		fseek(fp, 1024, SEEK_SET);		/* skip header */
		time_data = (unsigned long long int *) malloc(time_file_len);
		if (time_data == NULL) {
			printf("Could allocate enough memory --> exiting\n");
			return;
		}		
		n_read = fread(time_data, sizeof(unsigned long long int), (size_t) tot_time_entries, fp);
		if (n_read != tot_time_entries) {
			printf("Error reading time file --> exiting\n");
			return;
		}		
		fclose(fp);
		
		/* find sample index preceding requested range */
		for (i = 2; i < tot_time_entries; i += 3)
			if (time_data[i] > start_idx)
				break;
		curr_idx = time_data[i - 3];
		start_data_file_offset = time_data[i - 4];		/* start at key-sample flag */
		
		/* find sample index following requested range */
		for (; i < tot_time_entries; i += 3)
			if (time_data[i] > end_idx)
				break;
		if (i < tot_time_entries)
			end_data_file_offset = time_data[i - 1] - 1;	/* preceding byte is key-sample flag, so move back 1 */
		else
			end_data_file_offset = EOF;
		free(time_data);
		
		/* read in data file */
		fp = fopen(f_name, "r");
		if (fp == NULL) { 
			printf("Could not open the data file \"%s\" --> exiting\n",  f_name);
			return;
		}
		fseek(fp, 8, SEEK_SET);					/* read key sample flag value */
		fread(&temp_ks_flag, sizeof(short), (size_t) 1, fp);
		if (temp_ks_flag != -128) {
			fprintf(stderr, "\n\tUpdate software for unrecognized key-sample flag: %hd (0x%hX)\n\n", temp_ks_flag, temp_ks_flag);
			exit(1);
		}
		keysample_flag = (char) temp_ks_flag;
		
		if (end_data_file_offset == EOF) {
			fseek(fp, 0, SEEK_END);				/* get file length */
			end_data_file_offset = ftell(fp) - 1;		/*  offset of last byte of file */
		}
		comp_data_len = (end_data_file_offset - start_data_file_offset) + 1;
		comp_data = (char *) malloc(comp_data_len); 
		if (comp_data == NULL) {
			printf("Could allocate enough memory --> exiting\n");
			return;
		}		
		fseek(fp, start_data_file_offset, SEEK_SET);
		n_read = fread(comp_data, sizeof(char), (size_t) comp_data_len, fp);
		if (n_read != comp_data_len) {
			printf("Error reading data file --> exiting\n");
			return;
		}		
		fclose(fp);
		
		/* decompress */
		cdp = comp_data;
		data_end = comp_data + comp_data_len;
		cvp = (char *) &current_val;	
		for (j = curr_idx; j < start_idx; ++j) {		/* skip samples before requested start */
			if (*cdp == keysample_flag) {
				cvp[0] = (*++cdp < 0) ? -1 : 0;		/* sign extend */
				cvp[1] = *cdp++;
				cvp[2] = *cdp++;
				cvp[3] = *cdp++;
			} else
				current_val += (int) *cdp++;
		}
		for (k = 0; j <= end_idx; ++j) {			/* put samples in output array */
			if (*cdp == keysample_flag) {			
				cvp[0] = (*++cdp < 0) ? -1 : 0;		/* sign extend */
				cvp[1] = *cdp++;
				cvp[2] = *cdp++;
				cvp[3] = *cdp++;
			} else
				current_val += (int) *cdp++;
			decomp_arr[k++] = current_val;
		}
		
		free(comp_data);
		
		return;
		
	} else { /* endianness mismatch - THIS SECTION HAS NOT BEEN TESTED */
		
		/* read in time file */
		fseek(fp, 16, SEEK_SET);		/* number of time entries offset */
		fread(&temp_ui8, sizeof(unsigned long long int), 1, fp);
		b1 = (char *) &temp_ui8; b2 = b1 + 7;	/* byte-reverse number of time_entries */
		for (i = 8; i--;) {t = *b1; *b1++ = *b2; *b2-- = t;}
		n_time_entries = (int) temp_ui8;	/* if exceeds this, to big for Matlab anyway */
		tot_time_entries = n_time_entries * 3;	/* 3 fields per entry */
		fseek(fp, 0, SEEK_END);			/* get file length */
		time_file_len = ftell(fp) - 1024;
		fseek(fp, 1024, SEEK_SET);		/* skip header */
		time_data = (unsigned long long int *) malloc(time_file_len);
		if (time_data == NULL) {
			printf("Could allocate enough memory --> exiting\n");
			return;
		}		
		n_read = fread(time_data, sizeof(unsigned long long int), (size_t) tot_time_entries, fp);
		if (n_read != tot_time_entries) {
			printf("Error reading time file --> exiting\n");
			return;
		}		
		fclose(fp);
		for (i = 0; i < tot_time_entries; ++i) {		/* byte-reverse number of time_data array */
			b1 = (char *) (time_data + i) ; b2 = b1 + 7;
			for (k = 8; k--;) {t = *b1; *b1++ = *b2; *b2-- = t;}
		}
		
		/* find sample index preceding requested range */
		for (i = 2; i < tot_time_entries; i += 3)
			if (time_data[i] > start_idx)
				break;
		curr_idx = time_data[i - 3];
		start_data_file_offset = time_data[i - 4];		/* start at key-sample flag */
		
		/* find sample index following requested range */
		for (; i < tot_time_entries; i += 3)
			if (time_data[i] > end_idx)
				break;
		if (i < tot_time_entries)
			end_data_file_offset = time_data[i - 1] - 1;	/* preceding byte is key-sample flag, so move back 1 */
		else
			end_data_file_offset = EOF;
		free(time_data);
		
		/* read in data file */
		fp = fopen(f_name, "r");
		if (fp == NULL) { 
			printf("Could not open the data file \"%s\" --> exiting\n",  f_name);
			return;
		}
		fseek(fp, 8, SEEK_SET);					/* read key sample flag value */
		fread(&temp_ks_flag, sizeof(short), (size_t) 1, fp);
		b1 = (char *) &temp_ks_flag; b2 = b1 + 1;			/* byte-reverse key_sample flag value */
		for (i = 2; i--;) {t = *b1; *b1++ = *b2; *b2-- = t;}
		if (temp_ks_flag != -128) {
			fprintf(stderr, "\n\tUpdate software for unrecognized key-sample flag: %hd (0x%hX)\n\n", temp_ks_flag, temp_ks_flag);
			exit(1);
		}
		keysample_flag = (char) temp_ks_flag;
		
		if (end_data_file_offset == EOF) {
			fseek(fp, 0, SEEK_END);				/* get file length */
			end_data_file_offset = ftell(fp) - 1;		/* offset of last byte of file */
		}
		comp_data_len = (end_data_file_offset - start_data_file_offset) + 1;
		comp_data = (char *) malloc(comp_data_len); 
		if (comp_data == NULL) {
			printf("Could allocate enough memory --> exiting\n");
			return;
		}		
		fseek(fp, start_data_file_offset, SEEK_SET);
		n_read = fread(comp_data, sizeof(char), (size_t) comp_data_len, fp);
		if (n_read != comp_data_len) {
			printf("Error reading data file --> exiting\n");
			return;
		}		
		fclose(fp);
		
		/* decompress */
		if (file_endianness == BIG_ENDIAN_FILE) {
			cdp = comp_data;
			data_end = comp_data + comp_data_len;
			cvp = (char *) &current_val;	
			for (j = curr_idx; j < start_idx; ++j) {		/* skip samples before requested start */
				if (*cdp == keysample_flag) {
					cvp[3] = (*++cdp < 0) ? -1 : 0;		/* sign extend */
					cvp[2] = *cdp++;
					cvp[1] = *cdp++;
					cvp[0] = *cdp++;
				} else
					current_val += (int) *cdp++;
			}
			for (k = 0; j <= end_idx; ++j) {			/* put samples in output array */
				if (*cdp == keysample_flag) {			
					cvp[3] = (*++cdp < 0) ? -1 : 0;		/* sign extend */
					cvp[2] = *cdp++;
					cvp[1] = *cdp++;
					cvp[0] = *cdp++;
				} else
					current_val += (int) *cdp++;
				decomp_arr[k++] = current_val;
			}
		} else {
			cdp = comp_data;
			data_end = comp_data + comp_data_len;
			cvp = (char *) &current_val;	
			for (j = curr_idx; j < start_idx; ++j) {		/* skip samples before requested start */
				if (*cdp == keysample_flag) {
					cvp[3] = *++cdp;
					cvp[2] = *++cdp;
					cvp[1] = *++cdp;
					cvp[0] = (*cdp++ < 0) ? -1 : 0;		/* sign extend */
				} else
					current_val += (int) *cdp++;
			}
			for (k = 0; j <= end_idx; ++j) {			/* put samples in output array */
				if (*cdp == keysample_flag) {			
					cvp[3] = *++cdp;
					cvp[2] = *++cdp;
					cvp[1] = *++cdp;
					cvp[0] = (*cdp++ < 0) ? -1 : 0;		/* sign extend */
				} else
					current_val += (int) *cdp++;
				decomp_arr[k++] = current_val;
			}
		}
			
		free(comp_data);
		
		return;
	}	
}


/* The gateway routine */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	char			*f_name;
	int			buf_len, status, decomp_arr_len, dims[2], *decomp_arr;
	unsigned long long int	start_idx, end_idx;
	
	/*  Check for proper number of arguments */
	if (nrhs != 3) 
		mexErrMsgTxt("Three inputs required: file_name, start_index, stop_index");
	if (nlhs != 1) 
		mexErrMsgTxt("One output required: decompressed_array");
	

	/* Check to make sure the first input argument is a string */
	if (mxIsChar(prhs[0]) != 1)
		mexErrMsgTxt("Input must be a string.");
	
	/* Check to make sure the second input argument is a scalar */
	if( !mxIsDouble(prhs[1]) || mxIsComplex(prhs[1]) || (mxGetN(prhs[1]) * mxGetM(prhs[1]) != 1) ) {
		mexErrMsgTxt("Start index must be a scalar.");
	}	
	
	/* Check to make sure the third input argument is a scalar */
	if( !mxIsDouble(prhs[2]) || mxIsComplex(prhs[2]) || (mxGetN(prhs[2]) * mxGetM(prhs[2]) != 1) ) {
		mexErrMsgTxt("Start index must be a scalar.");
	}	
	
	/* Get the length of the input string */
	buf_len = (mxGetM(prhs[0]) * mxGetN(prhs[0])) + 2;
	
	/* Allocate memory for file_name string. */
	f_name = malloc(buf_len);
	
	/* Copy the string data from prhs[0] into a C string */
	status = mxGetString(prhs[0], f_name, buf_len);
	if (status != 0) 
		mexWarnMsgTxt("Not enough space. String is truncated.");
	
	/*  get the start index */
	start_idx = (unsigned long long int) mxGetScalar(prhs[1]);
	--start_idx;     /* convert to C indexing */
	
	/*  get the end index */
	end_idx = (unsigned long long int) mxGetScalar(prhs[2]);
	--end_idx;     /* convert to C indexing */

	/* Set the output pointer to the output matrix. */
	decomp_arr_len = (int) ((end_idx - start_idx) + 1);
	/*plhs[0] = mxCreateDoubleMatrix(decomp_arr_len, 1, mxREAL);*/
    plhs[0] = mxCreateNumericArray(2,dims,mxINT32_CLASS,mxREAL);
	dims[0] = decomp_arr_len; dims[1] = 1;
	plhs[0] = mxCreateNumericArray(2, dims, mxINT32_CLASS, mxREAL);
	
	/* Create a C pointer to a copy of the output matrix. */
	decomp_arr = (int *) mxGetPr(plhs[0]);
	
	/* Call the C subroutine. */
	decomp_mayo(f_name, start_idx, end_idx, decomp_arr);
}
